/*
COMMENTI:

utilizzo:
digita il file eseguibile senza parametri e leggi le istruzioni 


*/


# ifndef        STREAMING_H
#include "streaming.h"
# endif

/***************************
**  Function Definitions  **
***************************/

/*
**  Function  : void stat_var_definition(void)
**  Return    : none
**  Remarks   : defines the statistical parameters observed; for more details,
**              please refer to Statistics User Manual  
*/

void stat_var_definition(void)
{
int i=0;

name_var_stat[0] = "pkt_loss";
name_var_stat[1] = "pkt_delay";
name_var_stat[2] = "queue_dimension";
name_var_stat[3] = "queue_dimension_continuous_time";
name_var_stat[4] = "active_source";

  for (i = 0; i < N_STAT_VAR; i++)
    var_id[i] = i;

  for (i = 0; i < N_STAT_VAR; i++)
     f_vect1[i] = f_vect2[i] = NULL;
  
  /* for SIGNIFICANT variables, f_vect1 points to a function that resets some 
     performance parameters; this is done after transient period is over */    
  f_vect1[1] = reset; 
  f_vect1[3] = reset;
  
  /* assignement of SIGNIFICANT flag to statistical parameters  */
  significant[var_id[0]] = FALSE;  /* pkt_loss  */
  significant[var_id[1]] = TRUE;   /* pkt_delay */
  significant[var_id[2]] = FALSE;  /* queue_dimension */
  significant[var_id[3]] = FALSE;  /* queue_dimension_continuous_time */
  significant[var_id[4]] = FALSE;  /* active_source */
}


/*
**  Function  : void init_var(void)
**  Return    : none
**  Remarks   : this function initializes all the variables; it has to be
**              called after the input file was already read (so that
**              "param_collection" vector contanins meaningful values)  
*/
void init_var(void)
{
int index;

  current_time = 0.0;
  sched_events_cnt = 0;
  real_events_cnt = 0;
  
  num_active_nodes = (int)(initial_percentage_active_nodes*actual_percentage_arriving_nodes*max_num_peers);
  num_non_active_nodes = max_num_peers - num_active_nodes;
  cumulative_arrived  = num_active_nodes;
  cumulative_left  = 0;
  cumulative_error = 0;
  missed_join = 0;
  cumulative_orphan = 0;
  cumulative_solved = 0;     
  /******************************************************************/
  /*  creation of structures                                        */
  /******************************************************************/

  /* Here we create the vector of Node objects */
  node_list = malloc(max_num_peers*sizeof(Node*));
  non_active_nodes = malloc(max_num_peers*sizeof(Node*));
  
  /* this vector is used in case of uniform generation of the neighbors */
  neighbors_id_basic = malloc(max_num_neighbors*sizeof(int)); 
  /* value (determinitic) assignment */ 
  for (i=0; i < (max_num_neighbors/2); i++){
      dummy  = (int)pow(2, (i*2*8*sizeof(int)/(double)max_num_neighbors)); 
      neighbors_id_basic[2*i]   = dummy%num_active_nodes;  
      neighbors_id_basic[2*i+1] = -(dummy%num_active_nodes); 
  }
  
  /* instead of using "neighbors_id_basic" we create the adjacency matrix */
  adj_matrix = malloc(num_active_nodes*sizeof(int*));
  for(i = 0; i < num_active_nodes; i++){
      adj_matrix[i] = malloc((max_num_neighbors/2)*sizeof(int));
  }//for k     
  for(i = 0; i < num_active_nodes; i++){
     for(j = 0; j < (max_num_neighbors/2); j++){
      adj_matrix[i][j] = -1;
     }//for j     
  }//for i     

  create_adj_matrix(num_active_nodes, (max_num_neighbors/2), adj_matrix); 
  
  /* We initialize many things */
  for (i=0; i < max_num_peers; i++){
    node_list[i] = malloc(sizeof(Node));
    non_active_nodes[max_num_peers - 1 - i] = node_list[i];
    (*node_list[i]).level        = -1;
    (*node_list[i]).class        = -1;
    (*node_list[i]).index        = i;
    (*node_list[i]).dwl_bw       = 0.0;
    (*node_list[i]).upl_bw       = 0.0;
    (*node_list[i]).rate         = 0.0;
    (*node_list[i]).end_chunk    = 0.0;
    (*node_list[i]).last_change  = 0.0;
    (*node_list[i]).num_children = 0;
    (*node_list[i]).num_active_children = 0;
    if (i < num_active_nodes) 
         (*node_list[i]).leave_time   = HUGE_VAL; /* constant defined in math.h that means "infinity" */
    else (*node_list[i]).leave_time   = 0; 
    (*node_list[i]).num_fathers  = 0;  
    (*node_list[i]).num_active_fathers  = 0;  
    (*node_list[i]).fathers_id   = malloc(2*NUM_STRIPES*sizeof(Node*));   /* 2NUM_STRIPES should be        */
    (*node_list[i]).rate_of_each = calloc(2*NUM_STRIPES, sizeof(double)); /* replaced with the ratio       */
    (*node_list[i]).stripe_of_each = calloc(2*NUM_STRIPES, sizeof(int));  /* between uplink and downlink   */
    (*node_list[i]).delay_of_each = calloc(2*NUM_STRIPES, sizeof(double));  /* bandwdth of the slowest node  */  
    (*node_list[i]).stripe        = 0;
    (*node_list[i]).num_neighbors = 0;
    (*node_list[i]).neighbors     = malloc(max_num_neighbors*sizeof(Node*));
  
    /* bandwidth assignation for active nodes*/
    if (i < num_active_nodes) {
       rnd_value = uniform_0_1(&seed);
       extracted_dwl_bw = base_chunk_vect[1][0];
       extracted_upl_bw = base_chunk_vect[2][0];
       for (j=0; j<num_deltas_pmf-1; j++){
            if (rnd_value > base_chunk_vect[0][j]){
                extracted_dwl_bw = base_chunk_vect[1][j+1];
                extracted_upl_bw = base_chunk_vect[2][j+1];
            }//if
            else {break;}    
       }//for i
      
       (*node_list[i]).class = j;
       (*node_list[i]).dwl_bw = extracted_dwl_bw/bw_normalization;
       (*node_list[i]).upl_bw = extracted_upl_bw/bw_normalization;
       (*node_list[i]).spare_dwl_bw = extracted_dwl_bw/bw_normalization;
       (*node_list[i]).spare_upl_bw = extracted_upl_bw/bw_normalization;
    }
  }//for i
  
  /* neighbors assignment: it must be done here since some node_list[index] has no meaning during
     this phase */
  for (i=0; i < num_active_nodes; i++){
    
     /* this is the case where "neighbors_id_basic" is used 
     (*node_list[i]).num_neighbors = max_num_neighbors/2;
     for (j=0; j < max_num_neighbors/2; j++){
          index = (i + num_active_nodes + neighbors_id_basic[j])%num_active_nodes;
          (*node_list[i]).neighbors[j] = node_list[index];
     }
     shuffle(max_num_neighbors/2, (*node_list[i]).neighbors, &seed);
      end of "neighbors_id_basic" case */   
     
     /* this is the case where "adj_matrix" is used   */
     j=0;
     while ((adj_matrix[i][j] >= 0)&&(j<max_num_neighbors/2)) {
        index = adj_matrix[i][j];
        (*node_list[i]).neighbors[j] = node_list[index];
        j++;
     }
     (*node_list[i]).num_neighbors = j;
     shuffle(j, (*node_list[i]).neighbors, &seed);
     /* end of "adj_matrix" case */   
  
  }  


    
  /* We create a vector that will contain delay for all nodes: 
     it is necessary for the final sorting. 
     We keep track of different pahses, so we use a matrix*/
  stat_delay    = malloc(N_PARAM_STAT_DELAY*sizeof(double*));
  for(k = 0; k < N_PARAM_STAT_DELAY; k++){
      stat_delay[k]    = malloc(max_num_peers*sizeof(double));
  }//for k     
  /* We initialize it */
  for (i=0; i < N_PARAM_STAT_DELAY; i++){
      for (j=0; j < max_num_peers; j++){
           stat_delay[i][j]    = 0.0;
      }//for j
  }//for i

  delay_distrib = malloc(4*sizeof(double*));
  delay_collect = malloc(4*sizeof(double*));
  expansion_rates = 10.0;
  dim_delay_distrib = (int)(expansion_rates*log(max_num_peers)/log(2));
  delta_variation = dim_delay_distrib/2;
  for(k = 0; k < 4; k++){
      delay_distrib[k] = malloc(dim_delay_distrib*sizeof(double));
      delay_collect[k] = malloc(dim_delay_distrib*sizeof(double));
  }//for k     
  /* We initialize it */
  for (j=0; j < dim_delay_distrib; j++){
       delay_distrib[REF_DELAY][j]    = j*((int)(1/streaming_rate+0.5));
       delay_distrib[DIFFUSION][j]    = 0.0;
       delay_distrib[X_CONNECT][j]    = 0.0;
       delay_distrib[DELT_POST][j]    = 0.0;
       delay_collect[REF_DELAY][j]    = j*((int)(1/streaming_rate+0.5));
       delay_collect[DIFFUSION][j]    = 0.0;
       delay_collect[X_CONNECT][j]    = 0.0;
       delay_collect[DELT_POST][j]    = 0.0;
  }//for j
  
  stat_topology = malloc(2*sizeof(double*));
  topology_distrib = malloc(2*sizeof(double*));
  dim_stat_topology = MAX(2*max_num_children, 2*NUM_STRIPES);
  for(k = 0; k < 2; k++){
      stat_topology[k]    = malloc(dim_stat_topology*sizeof(double));
      topology_distrib[k] = malloc(dim_stat_topology*sizeof(double));
  }//for k     
  /* We initialize it */
  for (i=0; i < 2; i++){
      for (j=0; j < dim_stat_topology; j++){
           stat_topology[i][j]    = 0.0;
           topology_distrib[i][j] = 0.0;
      }//for j
  }//for i

  stat_rates = malloc(4*sizeof(double*));
  rates_distrib = malloc(4*sizeof(double*));
  expansion_rates = 10.0;
  dim_stat_rates = expansion_rates*(max_num_children+1);
  for(k = 0; k < 4; k++){
      stat_rates[k]    = malloc(dim_stat_rates*sizeof(double));
      rates_distrib[k] = malloc(dim_stat_rates*sizeof(double));
  }//for k     
  /* We initialize it */
  for (j=0; j < dim_stat_rates; j++){
       stat_rates[KIND_RATE][j]    = (double)j*streaming_rate/expansion_rates;
       stat_rates[DWNL_RATE][j]    = 0;
       stat_rates[KINDRATIO][j]    = (double)j/expansion_rates;
       stat_rates[UPL_DWL_R][j]    = 0;
       rates_distrib[KIND_RATE][j] = (double)j*streaming_rate/expansion_rates;
       rates_distrib[DWNL_RATE][j] = 0;
       rates_distrib[KINDRATIO][j] = (double)j/expansion_rates;
       rates_distrib[UPL_DWL_R][j] = 0;
  }//for j

  stat_dim_trees_sample = calloc(NUM_STRIPES, sizeof(double));
  stat_dim_trees = calloc(NUM_STRIPES, sizeof(double));
    
  /* In this structure we put 200 samples of the end file times; 
     we record the number of i-th reached node and for each round 
     three parameters: the end file time, the square of end file time
     (for the viariance calculation) and the number of 
     taken samples */
  if(max_num_peers < 200) num_samples = max_num_peers;   
  else num_samples = 200;
  delay_samples    = malloc(num_samples*sizeof(double*));
  for(k = 0; k < num_samples; k++){
      /* 3 measures (mean, variance and num_sample) for 3 parameters
         + 1 column for the num. of peers*/
      delay_samples[k] = malloc((3*N_PARAM_STAT_DELAY + 1)*sizeof(double));
  }//for k     
  /* We initialize it */
  dummy = (int)( ((double)max_num_peers) / ((double)num_samples) + 0.5);
  for (i = 0; i < num_samples; i++){
      delay_samples[num_samples - 1 - i][0] = max_num_peers - i*dummy - 1;
      for (j = 1; j < (3*N_PARAM_STAT_DELAY + 1); j++){
          delay_samples[num_samples - 1 - i][j] = 0.0;
      }//for j
  }//for i


  /********************************************************************/
  /* all the following variable are used when a node start picking 
     children until it is saturated */
  b_children_vect    = malloc((max_num_children)*sizeof(double));
  b_children_vect_u  = malloc((max_num_children)*sizeof(double));
  b_allocated        = malloc((max_num_children)*sizeof(double));
  b_notcompl         = malloc((max_num_children)*sizeof(int));
  for (i=0; i<(max_num_children); i++){
       b_children_vect[i]    = 0.0;
       b_children_vect_u[i]  = 0.0;
       b_allocated[i]        = 0.0;
       b_notcompl[i]         = 0;
  }//for i

  stat_num_peers = malloc(6*sizeof(double));
     stat_num_peers[TOTAL___JOIN] = 0.0;
     stat_num_peers[TOTAL__LEAVE] = 0.0;
     stat_num_peers[PERCEN_LEAVE] = 0.0;
     stat_num_peers[TOTAL_ACTIVE] = 0.0;
     stat_num_peers[TOTAL__ERROR] = 0.0;
     stat_num_peers[PERCEN_ERROR] = 0.0;
     stat_num_peers[MISSED__JOIN] = 0.0;
     stat_num_peers[ORPHAN_ACTIV] = 0.0;
     stat_num_peers[SOLVED__FAST] = 0.0;

  fathers_t_run   = malloc(num_deltas_pmf*sizeof(double*));
  fathers_t_final = malloc(num_deltas_pmf*sizeof(double*));
  fathers_t_index = malloc(num_deltas_pmf*sizeof(int));
  fathers_t_point = malloc(num_deltas_pmf*sizeof(int));
  for(k = 0; k < num_deltas_pmf; k++){
      fathers_t_run[k]    = malloc(((int)(MAX_TIME))*sizeof(double));
      fathers_t_final[k]  = malloc(((int)(MAX_TIME))*sizeof(double));
  }//for k     
  /* We initialize it */
  for (i=0; i < num_deltas_pmf; i++){
      for (j=0; j < MAX_TIME; j++){
           fathers_t_run[i][j]   = 0.0;
           fathers_t_final[i][j] = 0.0;
      }//for j
      fathers_t_index[i] = -1;
      fathers_t_point[i] = 0;
  }//for i     
}


/*
**  Function  : void set_stat_parameters(void)
**  Return    : none
**  Remarks   : this function initializes all the statitical variables; 
**              it has to be called after the input file was already 
**              read (so that "stat_param_collection" vector contanins 
**              meaningful values)  
*/
void set_stat_parameters(void)
{
 set_ref_accuracy(stat_param_collection.accuracy);
 set_confid_lev(stat_param_collection.confidence_level);
 set_max_coll_obs(stat_param_collection.max_collected_observation);
 set_min_transient_length(stat_param_collection.min_transient_length);
 set_tr_bat_size(stat_param_collection.transient_batch_size);
 set_tr_seq_len(stat_param_collection.transient_sequence_length);
 set_tr_accuracy(stat_param_collection.transient_accuracy);
 set_st_bat_size( stat_param_collection.stationary_batch_size);
}

/*
**  Function  : void usage(void)
**  Return    : none
**  Remarks   : print the utilization sintax
*/
void usage(void)
{
  printf("\nUsage: ./exe p1 p2 p3 p4 p5 p6 p7 p8 p9 p10 p11 p12\n");
  printf("Parameters:\n");
  printf("      p1 = input file name with pdf\n");
  printf("      p2 = output file radix \n");
  printf("      p3 = # max_iterations(accuracy)\n");
  printf("      p4 = max number of neighbors\n");
  printf("      p5 = min number of children \n");
  printf("      p6 = max number of children\n");
  printf("      p7 = number of stripes \n");
  printf("      p8 = percentage of initial active nodes\n");
  printf("      p9 = percentage of clip_time where arrivals concentrate\n");
  printf("      p10 = mean remaining time as multuple of clip_time\n");
  printf("      p11 = clip_time time for each run (e.g. 1000)\n");
  printf("      p12= exp in #peers = 10^(exp)\n");
  printf("      p13= percentage of stripes that composes the streaming (e.g. 0.8)\n");
  printf("      p14= (optional) another input file with some flags setted\n");
  printf("example: ./streaming_v05 dst0_3dd.txt output_radix 1000 60 1 8 8 0.1 0.2 5 1000 3 param_best_prefer_8.txt");
  exit(0);
}



/**********************************************************************/
/**********************************************************************/
/**                                                                  **/
/**                              MAIN                                **/
/**                                                                  **/
/**********************************************************************/
/**********************************************************************/

int main(int argc, char *argv[]){

 Event    *ev;
 Node *node_id, *from_node;
 double stripe_threshold_pct;
 Booltype dequeue, file_param;
 int additional;
/******************************************************************/
/*  read the input commands and register the parameters           */
/******************************************************************/
         
if(!((argc==13)||(argc==14))) {
  usage();
}

 namefilein        = argv[1];
 strcpy(out_file_radix, argv[2]);
 
 max_iterations    = atoi(argv[3]);
 max_num_neighbors = atoi(argv[4]);
 min_num_children  = atoi(argv[5]);
 max_num_children  = atoi(argv[6]);
 NUM_STRIPES       = atoi(argv[7]);
 initial_percentage_active_nodes        = atof(argv[8]);
 time_span_for_arrival                  = atof(argv[9]);
 p_churn           = atof(argv[10]);
 MAX_TIME          = atof(argv[11]);
 MAX_EXP           = atof(argv[12]); 
 stripe_threshold_pct   = atof(argv[13]); 
 
 file_param = FALSE;
 if (argc==15){
   file_param = TRUE;
   namefile_param    = argv[14];
 }
  
 MAX_LEVELS        = 100;
 chunks            = 100;

 drastic_change = FALSE;
 if ((max_iterations%2) == 1)
    drastic_change = TRUE;

  
 max_num_peers = (int)pow(10.0, MAX_EXP);
 actual_percentage_arriving_nodes = 0.5;
 max_num_peers /= actual_percentage_arriving_nodes;
 //max_num_peers = 20;
 
 seed = DEFAULT_SEED;

/* output file for statistics is opened                                                   */   
   strcpy(namefileout, out_file_radix);
   strcat(namefileout, ".vso");
   if((fp_statistics = fopen(namefileout, "w")) == NULL)
     {
      printf("\nError:  the file %s cannot be opened.\n", namefileout);
      exit(1);
     }
 
/* carico il file di input che contiene la pmf (delle bande); come risultato mi da' 
   anche il numero di delta che compongono tale pmf;
   ATTENZIONE: base_chunk_vect contiene gia' la cumulativa */

num_deltas_pmf = read_input(namefilein);

/*  bw_normalization contiene il fattore che normalizza il valore delle bande in 
    modo tale che il la matrice che contiene le distribuzioni finali sia
    sufficientemente ampia; i valori vengono scelti tra la banda di download */
bw_normalization = base_chunk_vect[1][0]*base_chunk_vect[1][num_deltas_pmf-1];

streaming_rate = base_chunk_vect[1][0]/bw_normalization;
stripe_threshold = (int)(stripe_threshold_pct * NUM_STRIPES);
stripe_rate = streaming_rate / stripe_threshold;


MAX_TIME = MAX_TIME/streaming_rate;

/* 1/streaming is the time to upload a single chunk to other nodes; 
   in fact, the dimesion of a single chunk is 1/NUM_STRIPES and the stripe rate
   is streaming_rate/NUM_STRIPES;  
   the lenght of the file (or observation time) is MAX_TIME times this interval
   MAX_TIME should be much greater than log(N);
   nodes leave the network with a probability p_churn and rate 'leave_rate'
   nodes arrives to the network with rate for now supposed constant, so that all
   the nodes arrive within MAX_TIME
*/

    init_var();

//diffusion_time_span = log(num_active_nodes)/streaming_rate;
diffusion_time_span = 2/streaming_rate;
time_span_for_arrival *= MAX_TIME;

leave_rate = 1.0/(p_churn*MAX_TIME);
arrival_rate = (max_num_peers*actual_percentage_arriving_nodes*(1-initial_percentage_active_nodes))/time_span_for_arrival;
arrival_rate += (max_num_peers*actual_percentage_arriving_nodes*initial_percentage_active_nodes)*leave_rate;


/********************************************************************/
/* all pthe parameters are now available: variable initialization */
//    init_var();

/* Initializing the Event list                                    */
    init_event_list();
    
/* Statistical routines initialization */
    stat_var_definition();
  
    init_statistics(N_STAT_VAR, significant, name_var_stat, 
                    f_vect1, f_vect2, stop_simulation, fp_statistics);

    set_stat_parameters();


/********************************************************************/
/********************************************************************/
/* now all the variables are allocated and initialized, then we can 
   proceed creating the trees                                       */
/********************************************************************/
/********************************************************************/

/* Obtaining the CPU time elapsed at the beginning of the main          */
/* simulation cycle.                                                    */
 start_CPU_time = clock();

 half_size_int = (8*sizeof(int)) >> 1;
 iterations = 0;
 dequeue = TRUE;
 
 if ((file_param)&&(parse_parameters(namefile_param))){
    printf("Parameters read from input file\n");
 }
 else {
    printf("No input file, default parameters\n");
   
   //attach_to_the_best = TRUE; 
   attach_to_the_best = FALSE; 
   let_diffusion = TRUE; 
   //threshold_for_fathers = NUM_STRIPES/2;
   threshold_for_fathers = NUM_STRIPES;
   prefer_newedge = TRUE;
   //prefer_newedge = FALSE;
   allow_stripe_change = TRUE;
   //allow_stripe_change = FALSE;
//   drastic_change = TRUE; //decrease the errors: switch stripe
   //drastic_change = FALSE;
 }
 

 /* Generates the first events                         */
 start_tree(NUM_STRIPES);
 if (arrival_rate > 0){
     new_arrival_time = diffusion_time_span + negexp((1/arrival_rate), &seed);
     schedule(GENERATION, new_arrival_time, NULL, NULL, 0);
 }
 
/*NOTE
ere we do not set the constraint on reached peers: in fact when all 
the peers are reached, no more events are generated and the simulator 
automatically reset */ 

#ifdef DEBUG
fprintf(stderr,"\n iterations:     0");
#endif

 
/* main simulation cycle                     */
  while (iterations < max_iterations)
    {
      ev           = head_of_event_list->next;
      current_time = ev->time;
      node_id      = ev->peer_id;
      from_node    = ev->from_peer;
      additional   = ev->param;
      ++real_events_cnt;
     
      /* if this run has reached the maximum time, then dequeue all the remaining events */         
      if (current_time > MAX_TIME) {
         while (ev->type != HEAD) {
               dequeue_event(ev);
               ev = head_of_event_list->next;
               ++real_events_cnt;
        }
      }
      
      switch (ev->type)
            {            
            case HEAD:
              dequeue = FALSE;       /* the event is not dequeued from the event list*/
              reset();
              start_tree(NUM_STRIPES);
              if (arrival_rate > 0){
                  new_arrival_time = diffusion_time_span + negexp((1/arrival_rate), &seed);
                  schedule(GENERATION, new_arrival_time, NULL, NULL, 0);
              }  
              iterations++;
              #ifdef DEBUG
                 fprintf(stderr,"\b\b\b\b\b");
                 fprintf(stderr,"%5d", iterations);
              #endif
            break;

            case JOIN:
              attach(node_id);
              dequeue = TRUE;            
            break;
            
            case GENERATION:
              generate_new_join();
              dequeue = TRUE;            
            break;
            
            case LEAVE:
              leave_network(node_id);
              dequeue = TRUE;            
            break;
            
            case ORPHAN:
              search_new_father(node_id, from_node, additional);
              dequeue = TRUE;            
            break;
            
            case CHECK_INDEGREE:
              update_indegree(node_id);
              dequeue = TRUE;            
            break;
            
            default:
              printf ("Error in Main routine\n");
              exit (1);
            }/*switch*/

    /* analized event is freed and put in free_event list */         
    if(dequeue) dequeue_event(ev);

    
    
    } /*while*/
    
    printf("\n");
    stop_simulation();
    return(0);

}

