/*
COMMENTI:

*/

# ifndef        STREAMING_H
#define         STREAMING_H
# endif

#define         DEBUG


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "stathead.h"

/***************************
**  Constant Definitions  **
***************************/

#ifndef TRUE
#define TRUE        (unsigned char)1
#endif

#ifndef FALSE
#define FALSE       (unsigned char)0
#endif

#define ERROR  -38808

#define DEFAULT_SEED  (long) 14123452
#define MAX(a, b) ((a) > (b) ? (a) : (b))
#define MIN(a, b) ((a) < (b) ? (a) : (b))

#define MAX_CHILDREN  20
#define MIN_PERCENTAGE_USED_BW  0.75

#define TOTAL___JOIN  0
#define TOTAL__LEAVE  1
#define PERCEN_LEAVE  2
#define TOTAL_ACTIVE  3
#define TOTAL__ERROR  4
#define PERCEN_ERROR  5
#define MISSED__JOIN  6
#define ORPHAN_ACTIV  7
#define SOLVED__FAST  8

#define N_PARAM_STAT_DELAY  3
#define DIFFUSION  0
#define X_CONNECT  1
#define DELT_POST  2
#define REF_DELAY  3

#define KIND_RATE  0
#define DWNL_RATE  1
#define KINDRATIO  2
#define UPL_DWL_R  3

#define FATHERS   0
#define CHILDREN  1

#define HEAD           255     /* event 'head of the list reached'                            */
#define REACH          1       /* event 'new peer reached'                                    */
#define CROSS_CONNECT  2       /* event 'create a cross connection'                           */
#define GENERATION     3       /* event 'node are generated by the generator'                 */
#define JOIN           4       /* event 'node enter in the network'                        */
#define LEAVE          5       /* event 'node leave the network'                              */
#define ORPHAN         6       /* event 'find another father for children of a leaving node'  */
#define DRASTIC_ATTACH 7       /* event 'deperately find another father'                      */
#define NOTIFY_CHANGE  8       /* event 'deperately find another father'                      */
#define CHECK_INDEGREE 9       /* event 'create a cross connection'                           */


#define BUNCHSIZE  10    /* number of simultaneously allocated events */
                         /* through malloc()                          */
#define N_STAT_VAR 5     /* number of monitored statistical variables */


/****************************
**  Structure Definitions  **
****************************/
 
/*
**  Structure : event_unit
**  Remarks   : collection of event attributes.
*/

typedef struct event_unit
{
  int type;                     /* event type                                       */
  struct event_unit *next;      /* pointer to the next record in the event list     */
  struct event_unit *prev;      /* pointer to the previous record in ev_list        */
  double             time;      /* occurence time of the event                      */
  struct node_unit  *peer_id;   /* pointer to the peer involved in the event        */
  struct node_unit  *from_peer; /* pointer to the peer that has generated the event */
  int  param;                   /* any other parameter */
}
Event;




typedef struct node_unit
{
  int                level;        /* it stores the number of level in the diffusion tree */
  int                class;        /* it stores the bandwdith class */
  int                index;        /* it stores the index of the reference vector */
  double             dwl_bw;       /* the download bandwidth of the node */
  double             upl_bw;       /* the upload bandwidth of the node */
  double             rate;         /* the actual rate used to transfer chunks*/
  double             end_chunk;    /* the instant when the first chunk is completely downloaded */
  double             last_change;  /* the instant when the last change in delay occured */
  int                num_children; /* number of children the node has */
  int                num_active_children; /* number of children the are really downloading from the node */
  double             spare_dwl_bw; /* actual spare download bandwidth */
  double             spare_upl_bw; /* actual spare upload bandwidth */
  double             leave_time;   /* time when the node will leave the network */  
  int                num_fathers;  /* number of nodes to download from */
  int                num_active_fathers;  /* number of fathers tfrom which the node is actually downloading */
  struct node_unit **fathers_id;   /* pointer to the list of node (to download from) identifiers */
  double            *rate_of_each; /* pointer to the list of rate received from each father */
  double            *delay_of_each; /* pointer to the list of delay of each father */
  int               *stripe_of_each; /* pointer to the list of stripe received from each father */
  int                stripe;       /* id of the stripes that have being downloaded */
  int                num_neighbors;  /* number of nodes to download from */
  struct node_unit **neighbors;    /* pointer to the list of neighbors */ 
}
Node;


/*
**  Structure : stat_param_unit
**  Remarks   : collection of statistical setting parameters
*/

typedef struct stat_param_unit
{
  double accuracy;                    /* for parameter definition, see      */ 
  int    confidence_level;            /* Statistics User Manual             */
  long   max_collected_observation;
  long   min_transient_length;
  long   transient_batch_size;
  long   transient_sequence_length;
  double transient_accuracy;
  long   stationary_batch_size;
}
Stat_Param;


/**********************************
**  Global Variable Definitions  **
**********************************/

double **base_chunk_vect;  


 /* system description */
 Node **node_list,       /* pointer to the list of the structure that contains node information */
      **non_active_nodes;/* pointer to the list of the structure that contains node information */
 int  **adj_matrix;      /* starting adjacency matrix */
 int max_num_neighbors,  /* maximum number of neighbors a node can have */
     max_num_peers,      /* maximum number of peers in the network */ 
     MAX_LEVELS,         /* define maximum nmber of level admitted for the tree */
     chunks,             /* num di chunk */
     num_deltas_pmf,     /* how many deltas are in the input pdf, i.e., how many classes */ 
     max_iterations,     /* number of iterations/sample of the markov process */ 
     max_num_children,   /* maximum outdegree */
     min_num_children,   /* minimum outdegree */
     bw_normalization,   /* used from bw to times, so that the maximum bandowdth becomes the maximum time*/
     NUM_STRIPES;        /* number of stripes the streaming is composed of */
 int *neighbors_id_basic;
 double MAX_EXP,         /* define the exponent such that the number of peers are 10^MAX_EXP */
        p_selfish,       /* probability to have a selfish peer that does not upload ot other nodes */  
        p_churn,         /* probability to have a peer that will leave the network */  
        arrival_rate,    /* join rate */  
        new_arrival_time,/* time of a new arrival */
        leave_rate,      /* leave rate */  
        stripe_rate,      /* rae of the single stripe */  
        streaming_rate;  /* rate of the streaming  */  
     
 /* system status */    
 long seed;
 
 int actual_levels,      /* stores temporarily the level of the node in the tree */
     num_active_nodes,     /* stores the number of active nodes */
     num_non_active_nodes, /* stores the number of non active nodes */
     cumulative_arrived,   /* stores the total number of arrived nodes */
     cumulative_left,      /* stores the total number of left nodes */
     cumulative_error,     /* stores the total number of nodes with error */
     missed_join,          /* a parameter that checks if a new request of join is not satisfied */
     cumulative_orphan,   /* stores the total number of orphan of active nodes */
     cumulative_solved,   /* stores the total number of orphan of active nodes immediatly solved */
     iterations,         /* it counts the number of current iterations (how many trees are built) */
     num_children,       /* actual number of children  */
     rounds,             /* it counts the number of current rounds */
     stripe_threshold,   /* it counts the minimum number of stripes necessary to receive the streaming */
     neighbor_rcv_stripe,  /* id of the stripe of the node */
     father_stripe;        /* id of the stripe of the father */
 double initial_percentage_active_nodes,     /* stores the percentage of initial active nodes */
        actual_percentage_arriving_nodes,
        b_father,        /* bw of the father node, used when it selects its children */
        b_father_transm, /* when the father is selecting its children, it register the actual rate used */
        b_children_tot,  /* when the father is selecting its children, it register their total bw */
        b_children_transm, /* it register the actual total rate used by children */ 
        used_bandwidth,    /* auxiliary parameter used in the process of children selection */ 
        extracted_dwl_bw, /* auxiliary parameter used to memorize dwl bw of the nodes */ 
        extracted_upl_bw;  /* auxiliary parameter used to memorize upl bw of the nodes */
 
 int     *b_notcompl,        /* auxiliary vector used in the bw allocation process */
         *b_children_id;     /* it registers the id of the first-generation children */
 double  *b_children_vect,   /* auxiliary vector used in the bw allocation process: dwl bw of the children */
         *b_children_vect_u, /* auxiliary vector used in the bw allocation process: upl bw of the children */ 
         *b_allocated;       /* auxiliary vector used in the bw allocation process: rate allocated */
 Booltype attach_to_the_best, let_diffusion, prefer_newedge, allow_stripe_change, drastic_change;
 int threshold_for_fathers;
 /* statistics */
 int num_samples;           /* number of samples taken at the end to display results */
 double **stat_delay,       /* matrix that stores the delay of each node */ 
        **delay_samples,    /* a sample of the previuos matrix */
        **delay_distrib,    /*  */
        **delay_collect,    /*  */
        **fathers_t_run,        /*  */
        **fathers_t_final,  /*  */
        **stat_rates,       /* matrix that stores the rate of each node */ 
        **rates_distrib,    /* a sample of the previuos matrix */
        **stat_topology,    /* matrix that stores some topology properties of each node */ 
        **topology_distrib, /* a sample of the previuos matrix */
         *stat_num_peers,   /* other measures on peers, such as the mean number of error nodes */
         *stat_dim_trees,   /* measures on number of peers per diffusion tree */
         *stat_dim_trees_sample,   /* measures on number of peers per diffusion tree */
         *histogram_node_50;/* store the histogram of the download times of the node with id = 50 */
 int     dim_delay_distrib, /* one dimension of the matrix "delay_distrib" */ 
         dim_stat_topology, /* one dimension of the matrix "stat_topology" */ 
         dim_stat_rates,    /* one dimension of the matrix "stat_rates" */ 
         *fathers_t_index,    /* vector that stores the indices of the observed nodes */ 
         *fathers_t_point,    /* vector that stores the value up to which the father_t_run is updated */ 
         delta_variation;   /* adjustment for delta delay */ 
 double  expansion_rates; 
 /* time variables */
 double   current_time,          /* absolute simulation time       */
          MAX_TIME,              /* maximum time, after which a single run ends: expressed as a mutiple of */
          time_span_for_arrival, /*     */
          diffusion_time_span;   /*     */

 clock_t  start_CPU_time;        /* CPU time elapsed at the beginning of */
                                /* the main simulation cycle.           */
 clock_t  end_CPU_time;          /* CPU time elapsed at the end of the   */
                                /* simulation run.                      */
/* Variables dealing with the management of the lists.*/
Event *head_of_event_list; /* pointer to the empty head of circular event list*/
Event *free_events;        /* pointer to the first element                   */
long     real_events_cnt;   /* Counter for the realized events  */
long     sched_events_cnt;  /* Counter for the scheduled events */         
                           /* in the free event list                         */
 
 /* auxiliary variables */
 FILE *fout;
 char *namefilein, namefileout[80], namefileout2[80], out_file_radix[80], *namefile_param;
 int i, j, k, dummy, aux_index, half_size_int;
 double cumulative_peers, cumulative_peers2, rnd_value, tmp, prev_value, prev_bw, total_bw;

 
Stat_Param   stat_param_collection;    /* structure that stores       */       
                                       /* statistical parameters      */       


/* Defines and initializes the variables used for the statistical	       */
/* analisys. The meaning and definition of each variable can be found in   */
/* the "Statistics User Manual".				                           */
FILE     *fp_statistics;         /* File collecting the statistical report */
int      n_stat_var;             /* number of monitored statistical var.   */
int      var_id[N_STAT_VAR];
void     (*f_vect1[N_STAT_VAR])();
void     (*f_vect2[N_STAT_VAR])();
Booltype significant[N_STAT_VAR];
char     *name_var_stat[N_STAT_VAR];


/***************************
**  Functions Prototypes  **
***************************/
 
/*  Module : neighbors_tv.c         */
/*  it contains 'main' routine */
void    stat_var_definition(void);
void    set_stat_parameters(void);
void    init_var (void);
void    usage(void);
 
/*  Module : random.c   */
long    rnd32(long);
double  uniform(double, double, long *);
double  uniform_0_1(long *);
double  negexp(double, long *);
int     poisson(double, long *);
int     geometric0(double, long *);
int     geometric1(double, long *);
int     geometric_trunc1(double, int, long *);
int     trunc_exp(double, long, long *);
 
/*  Module : lists.c    */
void    insert_event(Event *);
void    dequeue_event(Event *);
void    insert_free_events(Event *);
Event   *get_free_events(void);
Event   *new_event (void);
void    init_event_list(void);
void    dequeue_specific_event(int, double, Node *);

/*  Module : control.c    */
void    schedule(int, double, Node *, Node *, int);
void    start_tree(int);
void    node_reached(Node *, Node *);
void    create_cross_connections(Node *);
void    reset(void);
void    max_min_fairness_allocation(int, double, double *, double *, int *);
void    stop_simulation(void);
void    leave_network(Node *);
void    attach(Node *);
void    delay_changed(Node *, Node *);
void    generate_new_join(void);
void    search_new_father(Node *, Node *, int);
void    update_indegree(Node *);
int     choose_rnd_stripe(int );
void    create_adj_matrix(int, int, int **);


/*  Module : support.c    */
int     read_input(char *);
void    reset_stat (void);
void    shuffle(int, Node **, long *);
void    shuffle_int(int, int *, long *);
void    swap_ptr(Node **, int, int);
void    swap(double *, int, int);
void    swap_int(int *, int, int);
void    swap_int_2_vectors(int *, int, int, int *);
void    sort(double *, int, int);
void    sort_int(int *, int, int);
void    sort_int_2_vectors(int *, int, int, int *);
int      parse_line_traffic(char *, char *, char *);
Booltype parse_parameters(char *);


 

